__author__ = "Matthew Andres Moreno"
__copyright__ = "Copyright 2024, Matthew Andres Moreno"
__credits__ = ["Matthew Andres Moreno"]
__license__ = "MIT"
__version__ = "0.3.0"
__maintainer__ = "Matthew Andres Moreno"
__email__ = "m.more500@gmail.com"

from ._biopython_draw_colorclade_tree import biopython_draw_colorclade_tree
from ._draw_colorclade_tree import draw_colorclade_tree

__all__ = [
    "draw_colorclade_tree",
    "biopython_draw_colorclade_tree",
    "val_to_color_hls",
    "val_to_color_muted",
]



from collections import abc
import copy
import typing

from Bio import Phylo as BioPhylo
from Bio.Phylo.BaseTree import Clade as BioClade
from matplotlib import pyplot as plt
from matplotlib.axes import Axes as mpl_Axes

from ._biopython_color_tree import biopython_color_tree as color_tree
from ._biopython_pare_tree import biopython_pare_tree as pare_tree
from ._biopython_sort_tree import biopython_sort_tree as sort_tree
from ._matplotlib_drop_overlapping_labels import (
    matplotlib_drop_overlapping_labels,
)


def biopython_draw_colorclade_tree(
    tree: BioPhylo.BaseTree,
    ax: typing.Union[plt.Axes, tuple, None] = None,
    *,
    drop_overlapping_labels: bool = False,
    label_tips: typing.Union[typing.Callable[[BioClade], str], bool] = True,
    color_labels: typing.Union[typing.Callable[[str], str], str] = "black",
    line_width: float = 4.0,
    max_leaves: typing.Optional[int] = None,
    salt_color: typing.Optional[int] = None,
    use_branch_lengths: bool = True,
    val_to_color: typing.Optional[typing.Callable] = None,
) -> mpl_Axes:
    """Draw a phylogenetic tree from Biopython Tree, color-coding clades based
    on descendant labels.

    Labels are mapped to colors using a hash function. Color-coding uses an
    arbitrarily dominant color among the labels of the clade's descendants. The
    criteria used to determine dominance is deterministic, but arbitrary.

    Parameters
    ----------
    tree : BioPhylo.BaseTree
        The phylogenetic tree object to be visualized.
    ax : plt.Axes, tuple, or None, optional
        The matplotlib Axes object to draw on.

        If a tuple, it is interpreted as the size of the figure to create
        (width, height) in inches. If None, `plt.gca()` is used.
    drop_overlapping_labels : bool, default False
        If True, overlapping labels will be removed to improve clarity.
    label_tips : callable or bool, default True
        If a callable, then `label_tips` will be called on each tip node
        to determine if it should be labeled. If True, all tips of the tree
        will be labeled with their respective names.
    color_labels : callable or str, default "black"
        If a callable, then `color_labels` will be called on each tip name
        to determine what color it will be labeled with. If a string, it is
        interpreted directly as a color.
    line_width : float, default 4.0
        The width of the lines used to draw the tree.
    max_leaves : int or None, optional
        The maximum number of leaves to display.

        If the tree has more leaves, an arbitrary subset of the leaves will be
        displayed.
    salt_color : int or None, optional
        An optional integer to seed the random color generator for clade
        coloring.

        Useful to generate alternate colorings to find one that is visually
        appealing.
    use_branch_lengths : bool, default True
        If True, branch lengths are used to layout the tree. If False, tree is
        displayed as a cladogram.
    val_to_color : Callable or None, optional
        A function mapping each node's value to a color.

        If None, default coloring is applied.

    Returns
    -------
    mpl_Axes
        The matplotlib Axes object with the tree drawn.
    """
    if isinstance(label_tips, bool):
        # kwarg prevents unexpected late-binding behavior
        def label_tips(node: BioClade, label_tips=label_tips) -> str:
            return node.name or "" if label_tips else ""

    if isinstance(color_labels, str):
        # kwarg prevents unexpected late-binding behavior
        def color_labels(_: BioClade, color_labels=color_labels) -> str:
            return color_labels

    biopy_tree = copy.deepcopy(tree)
    if not use_branch_lengths:
        for clade in biopy_tree.find_clades(order="postorder"):
            clade.branch_length = 1.0

    if max_leaves is not None:
        biopy_tree = pare_tree(biopy_tree, max_leaves)

    sort_tree(biopy_tree.root)
    color_tree(biopy_tree.root, salt=salt_color, val_to_color=val_to_color)

    if isinstance(ax, abc.Sequence):
        _fig, ax = plt.subplots(figsize=ax, squeeze=True)
    elif ax is None:
        ax = plt.gca()
    elif not isinstance(ax, plt.Axes):
        raise TypeError(f"ax must be a Axes instance or tuple, not {type(ax)}")
    ax: plt.Axes  # ax has now been converted to a plt.Axes object

    with plt.rc_context({"lines.linewidth": line_width}):
        BioPhylo.draw(
            biopy_tree,
            axes=ax,
            label_func=lambda node: (
                label_tips(node) if node.is_terminal() else ""
            ),
            label_colors=color_labels,
            do_show=False,
        )

    if drop_overlapping_labels:
        matplotlib_drop_overlapping_labels(ax)

    # Remove axes borders except for bottom, and remove y-axis tick/labels
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.set_yticklabels([])
    ax.set_yticks([])
    ax.axes.get_yaxis().set_visible(False)

    return ax



import typing

from Bio import Phylo as BioPhylo
import opytional as opyt

from ._val_to_color_hls import val_to_color_hls


def biopython_color_tree(
    tree: BioPhylo.BaseTree,
    salt: typing.Optional[int] = None,
    val_to_color: typing.Optional[typing.Callable] = None,
) -> None:
    """Recursively color tree nodes based on dominant color among terminals."""
    if val_to_color is None:
        val_to_color = val_to_color_hls

    name = min(
        (opyt.or_value(y.name, "") for y in tree.get_terminals()),
        default=tree.name,
    )
    tree.color = val_to_color(name, salt=salt)

    for clade in tree.clades:
        biopython_color_tree(clade, salt=salt)



import colorsys
import random
import typing


def val_to_color_muted(
    val: typing.Any,
    salt: typing.Optional[int] = None,
    saturation: float = 1.0,
    brightness: float = 0.5,
) -> typing.Tuple[int, int, int]:
    """Converts a value to a color using a muted HSV mapping.

    Parameters
    ----------
    val : typing.Any
        The input value used to generate the color.

        This can be of any type that can be converted to a string.
    salt : typing.Optional[int], optional
        An optional integer used to salt the hash, providing a way to get
        different colors for the same input value.

    Returns
    -------
    typing.Tuple[int, int, int]
        A tuple representing the RGB color corresponding to the input value,
        with each component in the 0-255 range.
    """
    # Convert the hash to a value between 0 and 1
    rand = random.Random()
    rand.seed(str(val) + str(salt))
    hue = rand.random()

    # Convert the HSV color to an RGB color
    rgb_color = colorsys.hsv_to_rgb(hue, saturation, brightness)

    # Convert RGB values into 0-255 range (suitable for use in digital color)
    rgb_color_255 = tuple(int(c * 255) for c in rgb_color)

    return rgb_color_255



from Bio import Phylo as BioPhylo

from ._hash_name import hash_name


def biopython_pare_tree(
    tree: BioPhylo.BaseTree, max_leaves: int
) -> BioPhylo.BaseTree:
    """Trim the tree to keep only max_leaves leaves, selected based on the hash
    of their names."""

    # create a dictionary to store leaf names and their hashes
    hashes = {leaf.name: hash_name(leaf.name) for leaf in tree.get_terminals()}

    # sort the dictionary by hash value
    sorted_hashes = sorted(hashes.items(), key=lambda item: item[1])

    # keep only max_leaves leaves
    kept_names = {name for name, _ in sorted_hashes[:max_leaves]}

    # Traverse tree, but avoid modifying list during iteration
    while True:
        for clade in tree.find_clades(terminal=True):
            if clade.name not in kept_names:
                # If clade name isn't in kept_names, prune it
                tree.prune(clade)
                break
        else:
            break

    return tree



from matplotlib.axes import Axes as mpl_Axes


def matplotlib_drop_overlapping_labels(ax: mpl_Axes) -> None:
    """Iteratively hide labels from a matplotlib Axes object to resolve
    overlaps."""
    # Code to remove overlapping annotations
    bboxes = []
    for label in ax.texts:
        bbox = label.get_window_extent()
        # Check if current label overlaps with the others
        overlaps = any(bbox.overlaps(bbox_) for bbox_ in bboxes)
        if overlaps:
            label.set_visible(False)
        else:
            bboxes.append(bbox)



import typing

import alifedata_phyloinformatics_convert as apc
from matplotlib.axes import Axes as mpl_Axes
import pandas as pd

from ._biopython_draw_colorclade_tree import biopython_draw_colorclade_tree


def draw_colorclade_tree(
    tree: pd.DataFrame,
    backend: typing.Literal["biopython"],
    taxon_name_key: str,
    mutate: bool = False,
    *args,
    **kwargs,
) -> mpl_Axes:
    """Draw a phylogenetic tree given in alife standard as a pandas DataFrame,
    color-coding clades based on descendant labels.

    Labels are mapped to colors using a hash function. Color-coding uses an
    arbitrarily dominant color among the labels of the clade's descendants. The
    criteria used to determine dominance is deterministic, but arbitrary.

    Parameters
    ----------
    tree : pd.DataFrame
        The input phylogenetic tree data in a pandas DataFrame format.
    backend : typing.Literal["biopython"]
        The backend to use for drawing the phylogenetic tree.

        Currently, only 'biopython' is supported.
    taxon_name_key : str
        The column name in the DataFrame that contains the taxon names, which
        will be used as labels for the tree tips.
    mutate : bool, default False
        Can the input DataFrame be mutated?

        If False, a copy of the input DataFrame is made before drawing the tree.
    *args
        Variable length argument list, passed to the backend drawing function.
    **kwargs
        Arbitrary keyword arguments, passed to the backend drawing function.

    Returns
    -------
    mpl_Axes
        The matplotlib Axes object with the tree drawn.
    """
    if not mutate:
        tree = tree.copy()

    tree["name"] = tree[taxon_name_key]  # biopython expects "name" column

    rosetta_tree = apc.RosettaTree(tree)
    tree = {"biopython": lambda x: x.as_biopython}[backend](rosetta_tree)

    plotter = {"biopython": biopython_draw_colorclade_tree}[backend]
    return plotter(
        tree,
        *args,
        **kwargs,
    )



from Bio import Phylo as BioPhylo
import opytional as opyt


def biopython_sort_tree(tree: BioPhylo.BaseTree) -> None:
    """Recursively sort clades in-place in the tree based on their terminal
    elements' names."""
    tree.clades.sort(
        key=lambda x: min(
            (opyt.or_value(y.name, "") for y in x.get_terminals()),
            default=opyt.or_value(x.name, ""),
        )
    )

    for clade in tree.clades:
        biopython_sort_tree(clade)
